"""Implementation of Ewald processor class for a fixed size supercell.

An Ewald processor is optimized to compute electrostatic interaction energy
and changes in electrostatic energy from a list of local flips for use in
Monte Carlo sampling.

If you are using a Hamiltonian with a Cluster expansion and an Ewald summation
electrostatic term, you should use the CompositeProcessor with a
ClusterExpansionProcessor and an EwaldProcessor class.
"""

__author__ = "Luis Barroso-Luque"

import warnings
from functools import cached_property, lru_cache

import numpy as np
from pymatgen.analysis.ewald import EwaldSummation

from smol.cofe.extern.ewald import EwaldTerm
from smol.cofe.space.clusterspace import ClusterSubspace
from smol.correlations import delta_ewald_single_flip
from smol.moca.processor.base import Processor


class EwaldProcessor(Processor):
    """Processor for CE's including an EwaldTerm.

    A Processor class that handles changes for the electrostatic interaction
    energy using an Ewald Summation term.
    """

    def __init__(
        self,
        cluster_subspace,
        supercell_matrix,
        ewald_term,
        coefficient=1.0,
    ):
        """Initialize an EwaldProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                a cluster subspace.
            supercell_matrix (ndarray):
                an array representing the supercell matrix with respect to the
                Cluster Expansion prim structure.
            ewald_term (EwaldTerm):
                an instance of EwaldTerm to compute electrostatic energies.
            coefficient (float):
                Fitting coefficient to scale Ewald energy by.
        """
        contains_ewald = False
        for term in cluster_subspace.external_terms:
            if isinstance(term, EwaldTerm):
                contains_ewald = True
                break

        if not contains_ewald:
            cluster_subspace.add_external_term(EwaldTerm())
            warnings.warn(
                message="Warning: cluster subspace does not contain "
                "an Ewald term. Creating a default Ewald "
                "term and adding to cluster subspace"
            )
        super().__init__(cluster_subspace, supercell_matrix, coefficient)

        self._ewald_term = ewald_term
        # Set up ewald structure and indices
        struct, inds = self._ewald_term.get_ewald_structure(self.structure)
        self._ewald_structure = struct
        self._ewald_inds = np.ascontiguousarray(inds)

    @cached_property
    def ewald_summation(self):
        """Get the pymatgen EwaldSummation object."""
        ewald_summation = EwaldSummation(
            self._ewald_structure,
            real_space_cut=self._ewald_term.real_space_cut,
            recip_space_cut=self._ewald_term.recip_space_cut,
            eta=self._ewald_term.eta,
        )
        return ewald_summation

    @property
    @lru_cache(maxsize=None)
    def ewald_matrix(self):
        """Get the electrostatic interaction matrix.

        The matrix used is the one set in the EwaldTerm of the given
        ClusterExpansion.
        """
        matrix = self._ewald_term.get_ewald_matrix(self.ewald_summation)
        matrix = np.ascontiguousarray(matrix)
        return matrix

    def compute_property(self, occupancy):
        """Compute the Ewald electrostatic energy for a given occupancy array.

        Args:
            occupancy (ndarray):
                encoded occupancy array
        Returns:
            float: Ewald electrostatic energy
        """
        return self.coefs * self.compute_feature_vector(occupancy)

    def compute_property_change(self, occupancy, flips):
        """Compute change in electrostatic energy from a set of flips.

        Args:
            occupancy (ndarray):
                encoded occupancy array
            flips (list):
                list of tuples for (index of site, specie code to set)

        Returns:
            float: electrostatic energy change
        """
        return self.coefs * self.compute_feature_vector_change(occupancy, flips)

    def compute_feature_vector(self, occupancy):
        """Compute the feature vector for a given occupancy array.

        Each entry in the correlation vector corresponds to a particular
        symmetrically distinct bit ordering.

        Args:
            occupancy (ndarray):
                encoded occupation array

        Returns:
            array: correlation vector
        """
        ew_occu = self._ewald_term.get_ewald_occu(
            occupancy, self.ewald_matrix.shape[0], self._ewald_inds
        )
        return np.sum(self.ewald_matrix[ew_occu, :][:, ew_occu])

    def compute_feature_vector_change(self, occupancy, flips):
        """
        Compute the change in the feature vector from a list of flips.

        Args:
            occupancy (ndarray):
                encoded occupancy string
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a
                single flip where the first element is the index of the site
                in the occupancy string and the second element is the index
                for the new species to place at that site.

        Returns:
            array: change in correlation vector
        """
        occu_i = occupancy
        delta_energy = 0
        for f in flips:
            occu_f = occu_i.copy()
            occu_f[f[0]] = f[1]
            delta_energy += delta_ewald_single_flip(
                occu_f, occu_i, self.ewald_matrix, self._ewald_inds, f[0]
            )
            occu_i = occu_f
        return delta_energy

    def as_dict(self) -> dict:
        """
        Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        ewald_d = super().as_dict()
        ewald_d["ewald_summation"] = self.ewald_summation.as_dict()
        ewald_d["ewald_term"] = self._ewald_term.as_dict()
        return ewald_d

    @classmethod
    def from_dict(cls, d):
        """Create a EwaldProcessor from serialized MSONable dict."""
        # pylint: disable=duplicate-code
        proc = cls(
            ClusterSubspace.from_dict(d["cluster_subspace"]),
            np.array(d["supercell_matrix"]),
            ewald_term=EwaldTerm.from_dict(d["ewald_term"]),
            coefficient=d["coefficients"],
        )
        proc.ewald_summation = EwaldSummation.from_dict(d["ewald_summation"])

        return proc



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys
from datetime import date

import smol

sys.path.insert(0, os.path.abspath("../../smol"))
sys.path.insert(0, os.path.abspath("notebooks"))

# -- Project information -----------------------------------------------------

project = "smol"
copyright = f"2022-{date.today().year}, Ceder Group"
author = "Luis Barroso-Luque"

# The full version, including alpha/beta/rc tags
release = smol.__version__
version = smol.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    # "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    # "sphinx.ext.coverage",
    # "sphinx.ext.doctest",
    # "sphinx.ext.todo",
    # "sphinx.ext.viewcode",
    # "sphinx.ext.linkcode",
    "nbsphinx",
    "nb2plots",
    "IPython.sphinxext.ipython_console_highlighting",
]

# Generate the API documentation when building
autosummary_generate = True
add_module_names = False
autoclass_content = "both"

napoleon_google_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_custom_sections = None

# Add any paths that contain templates here, relative to this directory.
source_suffix = [".rst"]

# The encoding of src files.
source_encoding = "utf-8"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to src directory, that match files and
# directories to ignore when looking for src files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "pydata_sphinx_theme"

html_logo = (
    "../_static/logo.png"  # banner.svg needs text as paths to avoid font missing
)

html_theme_options = {
    "github_url": "https://github.com/CederGroupHub/smol",
    "use_edit_page_button": True,
    "show_toc_level": 2,
    # "navbar_align": "left",  # [left, content, right] For testing that the navbar
    # items align properly
    # "navbar_start": ["navbar-logo", "navbar-version"],
    # "navbar_center": ["navbar-nav", "navbar-version"],  # Just for testing
    "navigation_depth": 2,
    "show_nav_level": 2,
    "navbar_end": ["version-switcher", "navbar-icon-links"],  #
    # "left_sidebar_end": ["custom-template.html", "sidebar-ethical-ads.html"],
    # "footer_items": ["copyright", "sphinx-version", ""]
    "switcher": {
        # "json_url": "/_static/switcher.json",
        "json_url": "https://pydata-sphinx-theme.readthedocs.io/en/latest/_static/"
        "switcher.json",
        "url_template": "https://pydata-sphinx-theme.readthedocs.io/en/v{version}/",
        "version_match": version,
    },
    "external_links": [
        {
            "name": "Changes",
            "url": "https://github.com/CederGroupHub/smol/blob/master/CHANGES.md",
        },
        {"name": "Issues", "url": "https://github.com/CederGroupHub/smol/issues"},
    ],
}

html_context = {
    "github_url": "https://github.com",  # or your GitHub Enterprise interprise
    "github_user": "CederGroupHub",
    "github_repo": "smol",
    "github_version": "main",
    "doc_path": "docs/src",
    "source_suffix": source_suffix,
}

# Custom sidebar templates, maps page names to templates.
html_sidebars = {
    "contribute/index": [
        "search-field",
        "sidebar-nav-bs",
        "custom-template",
    ],  # This ensures we test for custom sidebars
    # "demo/no-sidebar": [],  # Test what page looks like with no sidebar items
}

# The style sheet to use for HTML and HTML Help pages. A file of that name
# must exist either in Sphinx' static/ path, or in one of the custom paths
# given in html_static_path.
# html_style = ''

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

html_css_files = [
    "css/smol.css",
]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Content template for the index page.
html_index = "index.html"

# If false, no module index is generated.
html_use_modindex = True

html_file_suffix = ".html"

# If true, the reST sources are included in the HTML build as _sources/<name>.
html_copy_source = False

# html_use_opensearch = "smolurl"

# Output file base name for HTML help builder.
htmlhelp_basename = "smol"

# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = r"""
{% set docname = 'docs/src/' + env.doc2path(env.docname, base=None) %}

.. raw:: html

    <div class="admonition note">
      This page was generated from
      <a class="reference external" href="https://github.com/CederGroupHub/smol/tree/{{ env.config.release|e }}/{{ docname|e }}">{{ docname|e }}</a>.<br>
      <script>
        if (document.location.host) {
          $(document.currentScript).replaceWith(
            '<a class="reference external" ' +
            'href="https://nbviewer.jupyter.org/url' +
            (window.location.protocol == 'https:' ? 's/' : '/') +
            window.location.host +
            window.location.pathname.slice(0, -4) +
            'ipynb">View & download notebook in <em>nbviewer</em></a>.'
          );
        }
      </script>
    </div>

.. raw:: latex

    \nbsphinxstartnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{The following section was generated from
    \sphinxcode{\sphinxupquote{\strut {{ docname | escape_latex }}}} \dotfill}}
"""



"""Classes for external terms that can be added to a cluster subspace.

Contains classes for external terms to be added to a cluster subspace
representing additional features to be fitted in a cluster expansion.

Currently only an Ewald electrostatic interaction term exists. Maybe it will
be the only term ever needed, but alas abstraction?
"""


from .ewald import EwaldTerm

__all__ = ["EwaldTerm"]



import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.analysis.structure_matcher import StructureMatcher

from smol.cofe.extern import EwaldTerm
from smol.cofe.space.domain import Vacancy, get_allowed_species
from smol.moca.processor import (
    ClusterExpansionProcessor,
    CompositeProcessor,
    EwaldProcessor,
)
from smol.moca.processor.base import Processor
from tests.utils import assert_msonable, gen_random_occupancy, gen_random_structure

RTOL = 0.0  # relative tolerance to check property change functions
# absolute tolerance to check property change functions (eps is approx 2E-16)
ATOL = 2e4 * np.finfo(float).eps
DRIFT_TOL = 10 * np.finfo(float).eps  # tolerance of average drift


@pytest.fixture
def ce_processor(cluster_subspace, rng):
    coefs = 2 * rng.random(cluster_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)
    return ClusterExpansionProcessor(
        cluster_subspace, supercell_matrix=scmatrix, coefficients=coefs
    )


@pytest.fixture(params=["real", "reciprocal", "point"])
def ewald_processor(cluster_subspace, rng, request):
    coef = rng.random(1)
    scmatrix = 3 * np.eye(3)
    ewald_term = EwaldTerm(use_term=request.param)
    return EwaldProcessor(
        cluster_subspace,
        supercell_matrix=scmatrix,
        coefficient=coef,
        ewald_term=ewald_term,
    )


# General tests for all processors
# Currently being done only on composites because I can not for the life of
# me figure out a clean way to parametrize with parametrized fixtures or use a
# fixture union from pytest_cases that works.
def test_encode_decode_property(composite_processor):
    occu = gen_random_occupancy(composite_processor.get_sublattices())
    decoccu = composite_processor.decode_occupancy(occu)
    for species, space in zip(decoccu, composite_processor.allowed_species):
        assert species in space
    npt.assert_equal(occu, composite_processor.encode_occupancy(decoccu))


def test_site_spaces(ce_processor):
    assert all(
        sp in ce_processor.structure.composition
        for space in ce_processor.unique_site_spaces
        for sp in space
        if sp != Vacancy()
    )
    assert all(
        sp in ce_processor._subspace.expansion_structure.composition
        for space in ce_processor.active_site_spaces
        for sp in space
        if sp != Vacancy()
    )


def test_sublattice(ce_processor):
    sublattices = ce_processor.get_sublattices()
    # These are default initialized, not splitted.
    site_species = get_allowed_species(ce_processor.structure)
    for sublatt, site_space in zip(sublattices, ce_processor.unique_site_spaces):
        assert sublatt.site_space == site_space
        for site in sublatt.sites:
            assert site_species[site] == list(site_space.keys())
    assert sum(len(sublatt.sites) for sublatt in sublattices) == len(
        ce_processor.structure
    )


def test_get_average_drift(composite_processor):
    forward, reverse = composite_processor.compute_average_drift()
    assert forward <= DRIFT_TOL and reverse <= DRIFT_TOL


def test_compute_property_change(composite_processor, rng):
    sublattices = composite_processor.get_sublattices()
    occu = gen_random_occupancy(sublattices)
    active_sublattices = [sublatt for sublatt in sublattices if sublatt.is_active]

    for _ in range(100):
        sublatt = rng.choice(active_sublattices)
        site = rng.choice(sublatt.sites)
        new_sp = rng.choice(sublatt.encoding)
        new_occu = occu.copy()
        new_occu[site] = new_sp
        prop_f = composite_processor.compute_property(new_occu)
        prop_i = composite_processor.compute_property(occu)
        dprop = composite_processor.compute_property_change(occu, [(site, new_sp)])
        # Check with some tight tolerances.
        npt.assert_allclose(dprop, prop_f - prop_i, rtol=RTOL, atol=ATOL)
        # Test reverse matches forward
        old_sp = occu[site]
        rdprop = composite_processor.compute_property_change(new_occu, [(site, old_sp)])
        assert dprop == -1 * rdprop


def test_structure_occupancy_conversion(ce_processor):
    sm = StructureMatcher()
    for _ in range(10):
        s_init = gen_random_structure(
            ce_processor.cluster_subspace.structure, size=ce_processor.supercell_matrix
        )
        s_init = s_init.get_sorted_structure()
        occu_init = ce_processor.occupancy_from_structure(s_init)

        s_conv = ce_processor.structure_from_occupancy(occu_init)
        s_conv = s_conv.get_sorted_structure()

        # occu_conv = ce_processor.occupancy_from_structure(s_conv)

        # For symetrically equivalent structures, StructureMatcher might generate
        # different structure_site_mappings
        # (see cluster_subspace.structure_site_mappings), therefore we may get
        # different occupancy strings with occupancy_from_structure, and
        # occu1 -> structure -> occu2 conversion cycle does not guarantee that
        # occu1 == occu2. In most use cases, it is not necessary to enforce that
        # occu1 == occu2. If you have to do so, you'll need to deeply modify the code of
        # StructureMatcher, which might not be a trivial task. Here we will only test
        # whether occu1 -> str1 and occu2 -> str2 are symetrically equivalent.
        # This should be enough in our application. We notify the users about this
        # mismatch in the documentations.
        assert sm.fit(s_init, s_conv)


def test_compute_feature_change(composite_processor, rng):
    sublattices = composite_processor.get_sublattices()
    occu = gen_random_occupancy(sublattices)
    active_sublattices = [sublatt for sublatt in sublattices if sublatt.is_active]
    composite_processor.cluster_subspace.change_site_bases("indicator")

    for _ in range(100):
        sublatt = rng.choice(active_sublattices)
        site = rng.choice(sublatt.sites)
        new_sp = rng.choice(sublatt.encoding)
        new_occu = occu.copy()
        new_occu[site] = new_sp
        prop_f = composite_processor.compute_property(new_occu)
        prop_i = composite_processor.compute_property(occu)
        dprop = composite_processor.compute_property_change(occu, [(site, new_sp)])
        # Check with some tight tolerances.
        npt.assert_allclose(dprop, prop_f - prop_i, rtol=RTOL, atol=ATOL)
        # Test reverse matches forward
        old_sp = occu[site]
        rdprop = composite_processor.compute_property_change(new_occu, [(site, old_sp)])
        assert dprop == -1 * rdprop


def test_compute_property(composite_processor):
    occu = gen_random_occupancy(composite_processor.get_sublattices())
    struct = composite_processor.structure_from_occupancy(occu)
    pred = np.dot(
        composite_processor.coefs,
        composite_processor.cluster_subspace.corr_from_structure(struct, False),
    )
    assert composite_processor.compute_property(occu) == pytest.approx(pred, abs=ATOL)


def test_msonable(composite_processor):
    occu = gen_random_occupancy(composite_processor.get_sublattices())
    d = composite_processor.as_dict()
    pr = Processor.from_dict(d)
    assert composite_processor.compute_property(occu) == pr.compute_property(occu)
    npt.assert_array_equal(composite_processor.coefs, pr.coefs)
    # send in pr bc composite_processor is scoped for function and new random
    # coefficients will be created.
    assert_msonable(pr)


# ClusterExpansionProcessor only tests
def test_compute_feature_vector(ce_processor):
    occu = gen_random_occupancy(ce_processor.get_sublattices())
    struct = ce_processor.structure_from_occupancy(occu)
    # same as normalize=False in corr_from_structure
    npt.assert_allclose(
        ce_processor.compute_feature_vector(occu) / ce_processor.size,
        ce_processor.cluster_subspace.corr_from_structure(struct),
    )


def test_bad_coef_length(cluster_subspace, rng):
    coefs = rng.random(cluster_subspace.num_corr_functions - 1)
    with pytest.raises(ValueError):
        ClusterExpansionProcessor(cluster_subspace, 5 * np.eye(3), coefficients=coefs)


def test_bad_composite(cluster_subspace, rng):
    coefs = 2 * rng.random(cluster_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)
    proc = CompositeProcessor(cluster_subspace, supercell_matrix=scmatrix)
    with pytest.raises(AttributeError):
        proc.add_processor(
            CompositeProcessor(cluster_subspace, supercell_matrix=scmatrix)
        )
    with pytest.raises(ValueError):
        proc.add_processor(
            ClusterExpansionProcessor(
                cluster_subspace, 2 * scmatrix, coefficients=coefs
            )
        )
    with pytest.raises(ValueError):
        new_cs = cluster_subspace.copy()
        ids = range(1, new_cs.num_corr_functions)
        new_cs.remove_corr_functions(rng.choice(ids, size=10))
        proc.add_processor(
            ClusterExpansionProcessor(new_cs, scmatrix, coefficients=coefs)
        )



# Copyright (c)

"""
smol -- Statistical Mechanics On Lattices

Lighthweight but caffeinated Python implementations of computational methods
for statistical mechanical calculations of configurational states in
crystalline material systems.


smol is a minimal implementation of computational methods to calculate statistical
mechanical and thermodynamic properties of crystalline material systems based on
the cluster expansion method from alloy theory and related methods. Although smol
is intentionally lightweight---in terms of dependencies and built-in functionality
---it has a modular design that closely follows underlying mathematical formalism
and provides useful abstractions to easily extend existing methods or implement and
test new ones. Finally, although conceived mainly for method development, smol can
(and is being) used in production for materials science research applications.
"""

import sys

from setuptools import Extension, dist, setup
from setuptools.command.build_ext import build_ext

# get numpy to include headers
dist.Distribution().fetch_build_eggs(["numpy>=1.20"])
import numpy

COMPILE_OPTIONS = {
    "msvc": [
        "/O2",
        "/EHsc",
    ],
    "mingw32": ["-Wall", "-Wextra"],
    "other": ["-Wall", "-Wextra", "-ffast-math", "-O3"],
}

LINK_OPTIONS = {
    "msvc": ["-Wl, --allow-multiple-definition"],
    "mingw32": [["-Wl, --allow-multiple-definition"]],
    "other": [],
}
COMPILER_DIRECTIVES = {
    "language_level": 3,
}


if sys.platform.startswith("darwin"):
    COMPILE_OPTIONS["other"] += ["-march=native", "-stdlib=libc++"]


# By subclassing build_extensions we have the actual compiler that will be used which is
# really known only after finalize_options
# http://stackoverflow.com/questions/724664/python-distutils-how-to-get-a-compiler-that-is-going-to-be-used  # noqa
class build_ext_options:
    def build_options(self):
        for e in self.extensions:
            e.extra_compile_args += COMPILE_OPTIONS.get(
                self.compiler.compiler_type, COMPILE_OPTIONS["other"]
            )
        for e in self.extensions:
            e.extra_link_args += LINK_OPTIONS.get(
                self.compiler.compiler_type, LINK_OPTIONS["other"]
            )


class build_ext_subclass(build_ext, build_ext_options):
    def build_extensions(self):
        build_ext_options.build_options(self)
        build_ext.build_extensions(self)


# Compile option for cython extensions
if "--use-cython" in sys.argv:
    USE_CYTHON = True
    cython_kwargs = {}
    sys.argv.remove("--use-cython")
    if "--annotate-cython" in sys.argv:
        cython_kwargs["annotate"] = True
        sys.argv.remove("--annotate-cython")
else:
    USE_CYTHON = False

ext = ".pyx" if USE_CYTHON else ".c"
ext_modules = [
    Extension("smol.correlations", ["smol/correlations" + ext], language="c")
]

if USE_CYTHON:
    from Cython.Build import cythonize

    ext_modules = cythonize(
        ext_modules,
        include_path=[numpy.get_include()],
        compiler_directives={"language_level": 3},
        **cython_kwargs
    )


setup(
    use_scm_version={"version_scheme": "python-simplified-semver"},
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext_subclass},
    include_dirs=numpy.get_include(),
)



"""Definitions of specific exceptions raised elsewhere."""


SYMMETRY_ERROR_MESSAGE = (
    "Error in calculating symmetry operations."
    "Try using a more symmetrically refined input"
    "structure. "
    "SpacegroupAnalyzer(s).get_refined_structure()"
    ".get_primitive_structure() "
    "usually results in a safe choice"
)


class NotFittedError(ValueError, AttributeError):
    """Exception if sklearn regression function is used before fitting.

    This class inherits from both ValueError and AttributeError to help with
    exception handling and backward compatibility.
    """


class SymmetryError(ValueError):
    """Exception for incompatibility between structure and given symops.

    Exception to raise when symmetry of a structure is not compatible with a
    set of given symops.
    """


class StructureMatchError(RuntimeError):
    """Raised when a pymatgen StructureMatcher returns None."""



"""Bias term definitions for biased sampling techniques.

Bias terms can be added to an MCKernel in order to generate samples that are
biased accordingly.
"""

__author__ = "Fengyu Xie, Luis Barroso-Luque"

from abc import ABC, abstractmethod
from collections import Counter
from math import log

import numpy as np

from smol.cofe.space.domain import get_species
from smol.utils import class_name_from_str, derived_class_factory


class MCBias(ABC):
    """Base bias term class.

    Attributes:
        sublattices (List[Sublattice]):
            list of sublattices with active sites.
    """

    def __init__(self, sublattices, rng=None, *args, **kwargs):
        """Initialize MCBias.

        Args:
            sublattices (List[Sublattice]):
                list of active sublattices, containing species information and
                site indices in sublattice.
            rng (np.Generator): optional
                The given PRNG must be the same instance as that used by the kernel and
                any bias terms, otherwise reproducibility will be compromised.
            args:
                Additional arguments buffer.
            kwargs:
                Additional keyword arguments buffer.
        """
        self.sublattices = sublattices
        self.active_sublattices = [
            sublatt for sublatt in self.sublattices if sublatt.is_active
        ]
        self._rng = np.random.default_rng(rng)

    @abstractmethod
    def compute_bias(self, occupancy):
        """Compute bias from occupancy.

        Args:
            occupancy(np.ndarray):
                encoded occupancy string.
        Returns:
            Float, bias value.
        """
        return

    @abstractmethod
    def compute_bias_change(self, occupancy, step):
        """Compute bias change from step.

        The returned value needs to be the difference of bias logs,
        log(bias_f) - log(bias_i), when the bias terms would directly multiply
        the ensemble probability (i.e. exp(-beta * E) * bias).

        Args:
            occupancy: (ndarray):
                encoded occupancy array.
            step: (List[tuple(int,int)]):
                step returned by MCUsher.
        Return:
            Float, change of bias value after step.
        """
        return


class FugacityBias(MCBias):
    """Fugacity fraction bias.

    This bias corresponds directly to using a composition bias. Using this
    with a CanonicalEnsemble keeps fugacity fractions constant, which
    implicitly sets the chemical potentials, albeit for a specific temperature.
    Since one species per sublattice is the reference species, to calculate
    actual fugacities, the reference fugacity must be computed as an ensemble
    average and all other fugacities can then be calculated.
    From the fugacities and the set temperature, the corresponding chemical
    potentials can then be calculated.
    """

    def __init__(self, sublattices, fugacity_fractions=None, **kwargs):
        """Initialize fugacity ratio bias.

        Args:
            sublattices (List[Sublattice]):
                list of active sublattices, containing species information and
                site indices in sublattice.
            fugacity_fractions (sequence of dicts): optional
                Dictionary of species name and fugacity fraction for each
                sublattice (.i.e think of it as the sublattice concentrations
                for random structure). If not given this will be taken from the
                prim structure used in the cluster subspace. Needs to be in
                the same order as the corresponding sublattice.
        """
        super().__init__(sublattices)
        self._fus = None
        self._fu_table = None
        # Consider only species on active sub-lattices
        self._species = [
            set(sublatt.site_space.keys()) for sublatt in self.active_sublattices
        ]

        if fugacity_fractions is not None:
            # check that species are valid
            fugacity_fractions = [
                {get_species(k): v for k, v in sub.items()}
                for sub in fugacity_fractions
            ]
        else:
            fugacity_fractions = [
                dict(sublatt.site_space) for sublatt in self.active_sublattices
            ]
        self.fugacity_fractions = fugacity_fractions

    @property
    def fugacity_fractions(self):
        """Get the fugacity fractions for species on active sublatts."""
        return self._fus

    @fugacity_fractions.setter
    def fugacity_fractions(self, value):
        """Set the fugacity fractions on active sublatts and update table."""
        for sub in value:
            for spec, count in Counter(map(get_species, sub.keys())).items():
                if count > 1:
                    raise ValueError(
                        f"{count} values of the fugacity for the same "
                        f"species {spec} were provided.\n Make sure the "
                        f"dictionaries you are using have only "
                        f"string keys or only Species objects as keys."
                    )

        value = [{get_species(k): v for k, v in sub.items()} for sub in value]
        if not all(sum(fus.values()) == 1 for fus in value):
            raise ValueError("Fugacity ratios must add to one.")
        for (spec, vals) in zip(self._species, value):
            if spec != set(vals.keys()):
                raise ValueError(
                    f"Fugacity fractions given are missing or not valid "
                    f"species.\n"
                    f"Values must be given for each  of the following: "
                    f"{self._species}"
                )
        self._fus = value
        self._fu_table = self._build_fu_table(value)

    def compute_bias(self, occupancy):
        """Compute bias from occupancy.

        Args:
            occupancy(np.ndarray):
                Encoded occupancy string.
        Returns:
            Float, bias value.
        """
        return sum(
            log(self._fu_table[site][species]) for site, species in enumerate(occupancy)
        )

    def compute_bias_change(self, occupancy, step):
        """Compute bias change from step.

        The returned value needs to be the difference of bias logs,
        log(bias_f) - log(bias_i), when the bias terms would directly multiply
        the ensemble probability (i.e. exp(-beta * E) * bias).

        Args:
            occupancy: (ndarray):
                encoded occupancy array.
            step: (List[tuple(int,int)]):
                step returned by MCUsher.
        Return:
            float, change of bias value after step.
        """
        delta_log_fu = sum(
            log(self._fu_table[f[0]][f[1]] / self._fu_table[f[0]][occupancy[f[0]]])
            for f in step
        )
        return delta_log_fu

    def _build_fu_table(self, fugacity_fractions):
        """Build an array for fugacity fractions for all sites in system.

        Rows represent sites and columns species. This allows quick evaluation
        of fugacity fraction changes from flips. Not that the total number
        of columns will be the number of species in the largest site space. For
        smaller site spaces the values at those rows are meaningless and will
        be given values of 1. Also rows representing sites with not partial
        occupancy will have all 1 values and should never be used.
        """
        num_cols = max(max(sublatt.encoding) for sublatt in self.sublattices) + 1
        # Sublattice can only be initialized as default, or splitted from default.
        num_rows = sum(len(sl.sites) for sl in self.sublattices)
        table = np.ones((num_rows, num_cols))
        for fus, sublatt in zip(fugacity_fractions, self.active_sublattices):
            ordered_fus = np.array([fus[sp] for sp in sublatt.site_space])
            table[sublatt.sites[:, None], sublatt.encoding] = ordered_fus[None, :]
        return table


def mcbias_factory(bias_type, sublattices, *args, **kwargs):
    """Get a MCMC bias from string name.

    Args:
        bias_type (str):
            string specifying bias name to instantiate.
        sublattices (List[Sublattice]):
            list of active sublattices, containing species information and
            site indices in sublattice.
        *args:
            positional args to instatiate a bias term.
        *kwargs:
            keyword argument to instantiate a bias term.
    """
    if "bias" not in bias_type and "Bias" not in bias_type:
        bias_type += "-bias"
    bias_name = class_name_from_str(bias_type)
    return derived_class_factory(bias_name, MCBias, sublattices, *args, **kwargs)



"""Implementation of MCMC transition kernel classes.

A kernel essentially is an implementation of the MCMC algorithm that is used
to generate states for sampling an MCMC chain.
"""

__author__ = "Luis Barroso-Luque"

from abc import ABC, abstractmethod
from math import log
from types import SimpleNamespace

import numpy as np

from smol.constants import kB
from smol.moca.sampler.bias import MCBias, mcbias_factory
from smol.moca.sampler.mcusher import MCUsher, mcusher_factory
from smol.utils import class_name_from_str, derived_class_factory, get_subclasses

ALL_MCUSHERS = list(get_subclasses(MCUsher).keys())
ALL_BIAS = list(get_subclasses(MCBias).keys())


class Trace(SimpleNamespace):
    """Simple Trace class.

    A Trace is a simple namespace to hold states and values to be recorded
    during MC sampling.
    """

    def __init__(self, /, **kwargs):  # noqa
        if not all(isinstance(val, np.ndarray) for val in kwargs.values()):
            raise TypeError("Trace only supports attributes of type ndarray.")
        super().__init__(**kwargs)

    @property
    def names(self):
        """Get all attribute names."""
        return tuple(self.__dict__.keys())

    def items(self):
        """Return generator for (name, attribute)."""
        yield from self.__dict__.items()

    def __setattr__(self, name, value):
        """Set only ndarrays as attributes."""
        if isinstance(value, (float, int)):
            value = np.array([value])

        if not isinstance(value, np.ndarray):
            raise TypeError("Trace only supports attributes of type ndarray.")
        self.__dict__[name] = value

    def as_dict(self):
        """Return copy of underlying dictionary."""
        return self.__dict__.copy()


class StepTrace(Trace):
    """StepTrace class.

    Same as Trace above but holds a default "delta_trace" inner trace to hold
    trace values that represent changes from previous values, to be handled
    similarly to delta_features and delta_energy.

    A StepTrace object is set as an MCKernel's attribute to record
    kernel specific values during sampling.
    """

    def __init__(self, /, **kwargs):  # noqa
        super().__init__(**kwargs)
        super(Trace, self).__setattr__("delta_trace", Trace())

    @property
    def names(self):
        """Get all field names. Removes delta_trace from field names."""
        return tuple(name for name in super().names if name != "delta_trace")

    def items(self):
        """Return generator for (name, attribute). Skips delta_trace."""
        for name, value in self.__dict__.items():
            if name == "delta_trace":
                continue
            yield name, value

    def __setattr__(self, name, value):
        """Set only ndarrays as attributes."""
        if name == "delta_trace":
            raise ValueError("Attribute name 'delta_trace' is reserved.")
        if not isinstance(value, np.ndarray):
            raise TypeError("Trace only supports attributes of type ndarray.")
        self.__dict__[name] = value

    def as_dict(self):
        """Return copy of serializable dictionary."""
        step_trace_d = self.__dict__.copy()
        step_trace_d["delta_trace"] = step_trace_d["delta_trace"].as_dict()
        return step_trace_d


class MCKernel(ABC):
    """Abtract base class for transition kernels.

    A kernel is used to implement a specific MC algorithm used to sample
    the ensemble classes. For an illustrative example of how to derive from this
    and write a specific kernel see the Metropolis kernel.
    """

    # Lists of valid helper classes, set these in derived kernels
    valid_mcushers = None
    valid_bias = None

    def __init__(
        self,
        ensemble,
        step_type,
        *args,
        seed=None,
        bias_type=None,
        bias_kwargs=None,
        **kwargs,
    ):
        """Initialize MCKernel.

        Args:
            ensemble (Ensemble):
                an Ensemble instance to obtain the features and parameters
                used in computing log probabilities.
            step_type (str):
                string specifying the MCMC step type.
            seed (int): optional
                non-negative integer to seed the PRNG
            bias_type (str): optional
                name for bias type instance.
            bias_kwargs (dict): optional
                dictionary of keyword arguments to pass to the bias
                constructor.
            args:
                positional arguments to instantiate the MCUsher for the
                corresponding step size.
            kwargs:
                keyword arguments to instantiate the MCUsher for the
                corresponding step size.
        """
        self.natural_params = ensemble.natural_parameters
        self._seed = seed if seed is not None else np.random.SeedSequence().entropy
        self._rng = np.random.default_rng(self._seed)
        self._compute_features = ensemble.compute_feature_vector
        self._feature_change = ensemble.compute_feature_vector_change

        self.trace = StepTrace(accepted=np.array([True]))
        self._usher, self._bias = None, None

        mcusher_name = class_name_from_str(step_type)
        self.mcusher = mcusher_factory(
            mcusher_name,
            ensemble.sublattices,
            *args,
            rng=self._rng,
            **kwargs,
        )

        if bias_type is not None:
            bias_name = class_name_from_str(bias_type)
            bias_kwargs = {} if bias_kwargs is None else bias_kwargs
            self.bias = mcbias_factory(
                bias_name,
                ensemble.sublattices,
                rng=self._rng,
                **bias_kwargs,
            )

        # run a initial step to populate trace values
        _ = self.single_step(np.zeros(ensemble.num_sites, dtype=int))

    @property
    def mcusher(self):
        """Get the MCUsher."""
        return self._usher

    @mcusher.setter
    def mcusher(self, usher):
        """Set the MCUsher."""
        if usher.__class__.__name__ not in self.valid_mcushers:
            raise ValueError(f"{type(usher)} is not a valid MCUsher for this kernel.")
        self._usher = usher

    @property
    def seed(self):
        """Get seed for PRNG."""
        return self._seed

    @property
    def bias(self):
        """Get the bias."""
        return self._bias

    @bias.setter
    def bias(self, bias):
        """Set the bias."""
        if bias.__class__.__name__ not in self.valid_bias:
            raise ValueError(f"{type(bias)} is not a valid MCBias for this kernel.")
        if "bias" not in self.trace.delta_trace.names:
            self.trace.delta_trace.bias = np.zeros(1)
        self._bias = bias

    def set_aux_state(self, state, *args, **kwargs):
        """Set the auxiliary state from initial or checkpoint values."""
        self._usher.set_aux_state(state, *args, **kwargs)

    @abstractmethod
    def single_step(self, occupancy):
        """Attempt an MCMC step.

        Returns the next state in the chain and if the attempted step was
        successful.

        Args:
            occupancy (ndarray):
                encoded occupancy.

        Returns:
            StepTrace: a step trace for states and traced values for a single
                       step
        """
        return self.trace

    def compute_initial_trace(self, occupancy):
        """Compute inital values for sample trace given an occupancy.

        Args:
            occupancy (ndarray):
                Initial occupancy

        Returns:
            Trace
        """
        trace = Trace()
        trace.occupancy = occupancy
        trace.features = self._compute_features(occupancy)
        # set scalar values into shape (1,) array for sampling consistency.
        trace.enthalpy = np.array([np.dot(self.natural_params, trace.features)])
        if self.bias is not None:
            trace.bias = np.array([self.bias.compute_bias(occupancy)])
        trace.accepted = np.array([True])
        return trace


class ThermalKernel(MCKernel):
    """Abtract base class for transition kernels with a set temperature.

    Basically all kernels should derive from this with the exception of those
    for multicanonical sampling and related methods
    """

    def __init__(self, ensemble, step_type, temperature, *args, **kwargs):
        """Initialize ThermalKernel.

        Args:
            ensemble (Ensemble):
                an Ensemble instance to obtain the features and parameters
                used in computing log probabilities.
            step_type (str):
                string specifying the MCMC step type.
            temperature (float):
                temperature at which the MCMC sampling will be carried out.
            args:
                positional arguments to instantiate the MCUsher for the
                corresponding step size.
            kwargs:
                keyword arguments to instantiate the MCUsher for the
                corresponding step size.
        """
        # hacky for initialization single_step to run
        self.beta = 1.0 / (kB * temperature)
        super().__init__(ensemble, step_type, *args, **kwargs)
        self.temperature = temperature

    @property
    def temperature(self):
        """Get the temperature of kernel."""
        return self.trace.temperature

    @temperature.setter
    def temperature(self, temperature):
        """Set the temperature and beta accordingly."""
        self.trace.temperature = np.array(temperature)
        self.beta = 1.0 / (kB * temperature)

    def compute_initial_trace(self, occupancy):
        """Compute inital values for sample trace given occupancy.

        Args:
            occupancy (ndarray):
                Initial occupancy

        Returns:
            Trace
        """
        trace = super().compute_initial_trace(occupancy)
        trace.temperature = self.trace.temperature
        return trace


class UniformlyRandom(MCKernel):
    """A Kernel that accepts all proposed steps.

    This kernel samples the random limit distribution where all states have the
    same probability (corresponding to an infinite temperature). If a
    bias is added then the corresponding distribution will be biased
    accordingly.
    """

    valid_mcushers = ALL_MCUSHERS
    valid_bias = ALL_BIAS

    def single_step(self, occupancy):
        """Attempt an MCMC step.

        Returns the next state in the chain and if the attempted step was
        successful.

        Args:
            occupancy (ndarray):
                encoded occupancy.

        Returns:
            StepTrace
        """
        step = self._usher.propose_step(occupancy)
        self.trace.delta_trace.features = self._feature_change(occupancy, step)
        self.trace.delta_trace.enthalpy = np.array(
            np.dot(self.natural_params, self.trace.delta_trace.features)
        )

        if self._bias is not None:
            self.trace.delta_trace.bias = np.array(
                self._bias.compute_bias_change(occupancy, step)
            )
            self.trace.accepted = np.array(
                True
                if self.trace.delta_trace.bias >= 0
                else self.trace.delta_trace.bias > log(self._rng.random())
            )

        if self.trace.accepted:
            for tup in step:
                occupancy[tup[0]] = tup[1]
            self._usher.update_aux_state(step)

        self.trace.occupancy = occupancy
        return self.trace


class Metropolis(ThermalKernel):
    """A Metropolis-Hastings kernel.

    The classic and nothing but the classic.
    """

    valid_mcushers = ALL_MCUSHERS
    valid_bias = ALL_BIAS

    def single_step(self, occupancy):
        """Attempt an MC step.

        Returns the next state in the chain and if the attempted step was
        successful.

        Args:
            occupancy (ndarray):
                encoded occupancy.

        Returns:
            StepTrace
        """
        step = self._usher.propose_step(occupancy)
        self.trace.delta_trace.features = self._feature_change(occupancy, step)
        self.trace.delta_trace.enthalpy = np.array(
            np.dot(self.natural_params, self.trace.delta_trace.features)
        )

        if self._bias is not None:
            self.trace.delta_trace.bias = np.array(
                self._bias.compute_bias_change(occupancy, step)
            )
            exponent = (
                -self.beta * self.trace.delta_trace.enthalpy
                + self.trace.delta_trace.bias
            )
            self.trace.accepted = np.array(
                True if exponent >= 0 else exponent > log(self._rng.random())
            )
        else:
            self.trace.accepted = np.array(
                True
                if self.trace.delta_trace.enthalpy <= 0
                else -self.beta * self.trace.delta_trace.enthalpy
                > log(self._rng.random())  # noqa
            )

        if self.trace.accepted:
            for tup in step:
                occupancy[tup[0]] = tup[1]
            self._usher.update_aux_state(step)
        self.trace.occupancy = occupancy

        return self.trace


def mckernel_factory(kernel_type, ensemble, step_type, *args, **kwargs):
    """Get a MCMC Kernel from string name.

    Args:
        kernel_type (str):
            string specifying kernel type to instantiate.
        ensemble (Ensemble)
            an Ensemble object to create the MCMC kernel from.
        step_type (str):
            string specifying the proposal type (i.e. key for MCUsher type)
        *args:
            positional arguments passed to class constructor
        **kwargs:
            keyword arguments passed to class constructor

    Returns:
        MCKernel: instance of derived class.
    """
    kernel_name = class_name_from_str(kernel_type)
    return derived_class_factory(
        kernel_name, MCKernel, ensemble, step_type, *args, **kwargs
    )



"""Implementation of SiteSpace and functions to get spaces and allowed species.

Implementation functions to generate lists of site spaces and allowed species
for a given disordered strucuture.

A site space is made up of the allowed species and their concentrations
including special species such as vacancies. A site spaces makes up the domain
of a site function and many site spaces make up the space of configurations.
"""

__author__ = "Luis Barroso-Luque, Fengyu Xie"

from collections import OrderedDict
from collections.abc import Hashable, Mapping

from monty.json import MSONable
from pymatgen.core import Composition
from pymatgen.core.periodic_table import DummySpecie, get_el_sp
from pymatgen.util.string import formula_double_format


def get_allowed_species(structure):
    """Get the allowed species for each site in a disordered structure.

    This will get all the allowed species for each site in a pymatgen structure.
    If the site composition does not add to 1, a Vacancy will be appended to
    the allowed species at that site.

    Same as in a site space, the order of the allowed species at each site is
    important because that implicitly represents the code that will be used
    when evaluating site basis functions.

    Args:
        structure (Structure):
            Structure to determine site spaces from. At least some sites should
            be disordered, otherwise there is no point in using this.

    Returns:
        list of Specie: list of allowed Species for each site
    """
    all_site_spaces = get_site_spaces(structure)
    return [list(site_space.keys()) for site_space in all_site_spaces]


def get_site_spaces(structure, include_measure=False):
    """Get site spaces for each site in a disordered structure.

    Method to obtain the single site spaces for the sites in a structure.
    The single site spaces are represented by the allowed species for each site
    and the measure/concentration for disordered sites. The order is important so
    the species need to be sorted, since that implicitly sets the code/index
    used to evaluate basis functions.

    Vacancies are included in sites where the site element composition does not
    sum to 1 (i.e. the total occupation is not 1).

    Args:
        structure (Structure):
            Structure to determine site spaces from. At least some sites should
            be disordered, otherwise there is no point in using this.
        include_measure (bool): optional
            if True, will take the site compositions as the site space measure,
            otherwise a uniform measure is assumed.

    Returns:
        SiteSpace: Allowed species and their corresponding measure.
    """
    all_site_spaces = []
    for site in structure:
        if include_measure:
            site_space = SiteSpace(site.species)
        else:  # make comp uniform if measure not included
            num_species = len(site.species)
            if site.species.num_atoms < 0.99:
                num_species += 1
            site_space = SiteSpace(
                Composition({sp: 1.0 / num_species for sp in site.species.keys()})
            )
        all_site_spaces.append(site_space)
    return all_site_spaces


def get_species(obj):
    """Get specie object.

    Wraps pymatgen.core.periodic_table.get_el_sp to be able to catch and
    return Vacancies when needed.

    Args:
        obj (Element/Specie/str/int):
            an arbitrary object.  Supported objects
            are actual Element/Specie objects, integers (representing atomic
            numbers) or strings (element symbols or species strings).

    Returns:
        Specie, Element or Vacancy, with a bias for the maximum number of
        properties that can be determined.
    Raises:
        ValueError if obj cannot be converted into an Element or Specie.
    """
    if isinstance(obj, Vacancy):
        return obj

    if isinstance(obj, (list, tuple)):
        return [get_species(o) for o in obj]

    if isinstance(obj, str) and "vac" in obj.lower():
        return Vacancy()

    return get_el_sp(obj)


class SiteSpace(Mapping, Hashable, MSONable):
    """Represents a site space.

    A site spaces is a set of allowed species at a site and their corresponding
    measure (composition in the disordered state).

    This class is somewhat similar to pymatgen Composition. However a SiteSpace
    must always have a total composition = 1 since vacancies are explicitly
    included as species. A sorted order of the included species is also kept to
    ensure a consistent encoding when creating basis functions. Vacancies are
    always the last species. Key errors are raised for species not in the
    site space.
    """

    def __init__(self, composition):
        """Initialize a site space.

        Args:
            composition (Composition):
                Composition object that specifies the site space.
        """
        # num_atoms = 0 identified as vacancy-only site-space.
        if composition.num_atoms < 0 or composition.num_atoms > 1:
            raise ValueError(
                "Number of atoms in Composition must be in [0, 1]"
                f" got {composition.num_atoms}!"
            )

        if any(isinstance(sp, Vacancy) for sp in composition):
            if composition.num_atoms != 1:
                raise ValueError(
                    f"Composition {composition} has a Vacancy but"
                    " number of atoms is not 1."
                )

            if sum(isinstance(sp, Vacancy) for sp in composition) > 1:
                raise ValueError(
                    "There are multiple vacancies in this composition "
                    f"{composition}."
                )

        self._composition = composition
        self._data = OrderedDict(
            (spec, comp) for spec, comp in sorted(composition.items())
        )
        if composition.num_atoms < 0.99:
            self._data[Vacancy()] = 1 - composition.num_atoms

    @property
    def composition(self):
        """Return the underlying composition."""
        return self._composition

    @property
    def codes(self):
        """Return range of species."""
        return tuple(range(len(self)))

    def __getitem__(self, item):
        """Get a specie in the sitespace."""
        try:
            specie = get_species(item)
            return self._data[specie]
        except ValueError as exp:
            raise TypeError(
                f"Invalid key {item}, {type(item)} for Composition"
                f".\nValueError exception:\n{exp}"
            ) from exp

    def __len__(self):
        """Get number os species."""
        return len(self._data)

    def __iter__(self):
        """Return OrderedDict iter."""
        return self._data.keys().__iter__()

    def __eq__(self, other):
        """Compare equality, requires same order."""
        if not isinstance(other, SiteSpace):
            return False

        return all(
            item1 == item2 for item1, item2 in zip(self._data.items(), other.items())
        )

    def __hash__(self):
        """Take the hash of the composition for now."""
        return hash(self.composition)

    def __str__(self):
        """Get pretty string."""
        return "\n".join(["Site Space", repr(self)])

    def __repr__(self):
        """Get repr."""
        out = str(self._composition) + " "
        if len(self._data) > len(self._composition):
            items = list(self._data.items())[len(self._composition) :]
            out += " ".join(
                [
                    f"{str(k)}{formula_double_format(v, ignore_ones=False)}"
                    for k, v in items
                ]
            )
        return out

    def as_dict(self) -> dict:
        """Get MSONable dict representation."""
        site_space_d = super().as_dict()
        site_space_d["composition"] = self._composition.as_dict()
        return site_space_d

    @classmethod
    def from_dict(cls, d):
        """Create a SiteSpace from dict representation."""
        return cls(Composition.from_dict(d["composition"]))


class Vacancy(DummySpecie):
    """Wrapper class around DummySpecie to treat vacancies as their own species."""

    def __init__(
        self, symbol: str = "A", oxidation_state: float = 0, properties: dict = None
    ):
        """Initialize a Vacancy.

        Args:
            symbol (str): an assigned symbol for the vacancy. Strict
                rules are applied to the choice of the symbol. The vacancy
                symbol cannot have any part of first two letters that will
                constitute an Element symbol. Otherwise, a composition may
                be parsed wrongly. E.g., "X" is fine, but "Vac" is not
                because Vac contains V, a valid Element.
            oxidation_state (float): oxidation state for Vacancy. More like
                the charge of a point defect. Defaults to zero.
        """
        super().__init__(
            symbol=symbol, oxidation_state=oxidation_state, properties=properties
        )

    def __eq__(self, other):
        """Test equality, only equal if isinstance."""
        return False if not isinstance(other, Vacancy) else super().__eq__(other)

    def __hash__(self):
        """Get hash, append an underscore to avoid clash with Dummy."""
        # this should prevent any hash collisions with other Specie instances
        # since the symbol for a DummySpecie can not start with "v"
        return hash("v" + self.symbol)

    def __str__(self):
        """Get string representation, add a v to differentiate."""
        return "vac" + super().__str__()

    def __repr__(self):
        """Get an explicit representation."""
        return "Vacancy " + self.__str__()

    def __copy__(self):
        """Copy the vacancy object."""
        return Vacancy(self.symbol, self._oxi_state, self._properties)

    def __deepcopy__(self, memo):
        """Deepcopy the vacancy object."""
        return Vacancy(self.symbol, self._oxi_state, self._properties)
